# Redux Polling
> Convenient way to support polling in your Redux app so you can focus on the business logic

[![NPM Version][npm-image]][npm-url]
[![Build Status][travis-image]][travis-url]
[![Coverage Status][coveralls-image]][coveralls-url]
<!-- [![Downloads Stats][npm-downloads]][npm-url] -->

Almost every app contains charts, progress bars or components that consist on repeated data sampling. This is done by Polling. Support polling in React-Redux application can be frustrating. Each polling scenario requires to write a dedicated reducer and actions, implement the polling (all the setTimeout / setInterval stuff), deal with history and this is even before writing the business logic of polling itself. 
Redux Polling provides a convenient way to support polling in React-Redux application so you will be able to focus on the business logic right away instead of dealing with all the boilerplate code. The setup is done by adding a middleware and you are ready to go.

## Installation

```sh
npm install --save redux-polling
```
Or
```sh
yarn add redux-polling
```

## Setup

1. Add `reduxPollingMiddleware` middleware to your Redux store:
```javascript store.js
import { createStore, applyMiddleware, compose } from 'redux';
import { reduxPollingMiddleware } from 'redux-polling';
import rootReducer from './path/to/your/root/reducer';

const initialState = {};
const middleware = [
    reduxPollingMiddleware,
];

const composedEnhancers = compose(
    applyMiddleware(...middleware),
);

const store = createStore(
    rootReducer,
    initialState,
    composedEnhancers,
);

export default store;
```

2. Add `reduxPollingReducer` reducer to your root reducer:
```javascript reducers.js
import { combineReducers } from 'redux';
import { reduxPollingNamespace, reduxPollingReducer } from 'redux-polling';

export default combineReducers({
    .....
    .....
    [reduxPollingNamespace]: reduxPollingReducer,
});
```

3. Your application is ready to use Redux Polling everywhere.

## Usage example

Let's say we want to implement a chart page that consist on data polling. The page has two components:
1. Action panel component that contains two buttons: "Start Polling" and "Stop Polling".
2. Chart component that shows the data.
   
Redux Polling provides everything for those components: start and stop actions for the buttons and selectors for receiving the results and history. All we left to do is to implement the polling logic that fetches the data once (Redux Polling will call it on each interval):

```javascript point-polling.js
import { createPollingActions, getPollingState } from 'redux-polling';
import { createSelector } from 'reselect';
import fetchNextPoint from './path/to/your/points/service';

/* Actions */
const pollingInterval = 1000;
const historyLength = 30;

async function polling() {
    const point = await fetchNextPoint();
    return point;
}

export const actions = createPollingActions('pointPolling', polling, pollingInterval, historyLength);

/* Selectors */
const isPollingActive = state => getPollingState(state, 'pointPolling').isActive;
const getPointHistory = state => getPollingState(state, 'pointPolling').history;
const getLastPoint = state => getPollingState(state, 'pointPolling').lastEntry;

export const selectors = {
    isPollingActive, getPointHistory, getLastPoint,
};
```

This module exports two items: `actions` and `selectors`.   
`actions` is an object that contains two action creators: `start()` and `stop()`.   
`selectors` is an object with selectors that our app needs.   
Please notice the use of `createPollingActions()` and `getPollingState()`.

That's all for the polling implementation. Now we can create the components.
1. Actions Panel:
```javascript actions-panel.jsx
import React from 'react';
import PropTypes from 'prop-types';
import { connect } from 'react-redux';
import { actions as pointPollingActions } from './state/point-polling';

function ActionsPanel(props) {
    const { startPolling, stopPolling } = props;
    return (
        <div>
            <button type="button" onClick={ () => startPolling() }>
                Start Polling
            </button>
            <button type="button" onClick={ () => stopPolling() }>
                Stop Polling
            </button>
        </div>
    );
}

ActionsPanel.propTypes = {
    startPolling: PropTypes.func,
    stopPolling: PropTypes.func,
};

ActionsPanel.defaultProps = {
    startPolling: () => {},
    stopPolling: () => {},
};

const mapDispatchToProps = {
    startPolling: pointPollingActions.start,
    stopPolling: pointPollingActions.stop,
};

export default connect(
    undefined,
    mapDispatchToProps,
)(ActionsPanel);
```
2. Chart (I will not get into implementation details of the chart since this is not the topic here):
```javascript chart.jsx
import React from 'react';
import PropTypes from 'prop-types';
import { connect } from 'react-redux';
import LineChart from 'my-favorite-charts-library';
import { selectors as pointPollingSelectors } from './state/point-polling';

function Chart(props) {
    const { isPollingActive, lastPoint, history } = props;
    return (
        <div>
            <div>
                { isPollingActive ? 'Polling Active' : 'Polling Inactive' }
            </div>
            <div>
                { `Last fetched point is ${lastPoint}` }
            </div>

            <LineChart data={ history } />
        </div>
    );
}

Chart.propTypes = {
    isPollingActive: PropTypes.bool,
    lastPoint: PropTypes.number,
    history: PropTypes.arrayOf(PropTypes.number),
};

Chart.defaultProps = {
    isPollingActive: false,
    lastPoint: undefined,
    history: [],
};

const mapStateToProps = state => ({
    isPollingActive: pointPollingSelectors.isPollingActive(state),
    lastPoint: pointPollingSelectors.getLastPoint(state),
    history: pointPollingSelectors.getPointHistory(state),
});

export default connect(
    mapStateToProps,
)(Chart);
```

Cool.



Example: https://naorye.github.io/redux-polling/
Notice that `polling()` can return either value or a promise.   


## Documentation
TBD

The main method is `createResource(method, apiUrl, options)` and it expects the following:
* `method` **[String](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String)** Can be one of `post`, `put`, `get` or `delete`
* `apiUrl` **[String](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String)** Api url (for example: `https://bittrex.com/api/v1.1/public/getmarkethistory`). See below for more info
* `options` - **[Object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object)** Supports `withCredentials`, `interpolationPattern`, `headersMap`, `inputMap` and `parsers`. See below for more info


### Url Interpolation

Sometimes we need to inject parameters to the api url. For example `GET https://api.example.com/chat/5/members` will be used to get the members list of room with id equal to `5`. Let's define such resource and use it:
```javascript
import { createResource } from 'plain-api';

const fetchChatMembers = createResource('get', 'https://api.example.com/chat/{{chatId}}/members');
...
...
...
const chatId = 5;
const members = await fetchChatMembers.call({ chatId });
```
`{{chatId}}` in the url is used as a placeholer. When calling the resource with `chatId = 5`, the parameter injected into the url.   
If we call the resource without providing the required interpolation params, the placeholders won't be replaced.

### Changing Interpolation Pattern

As a default, the regular expression that is used for injecting url parameters is `/\{\{(\w+)\}\}/gi` (which matches to all the wordes wraped inside `{{}}`). You can override this default by calling `setDefaultInterpolationPattern(pattern)`:   
```javascript
import { createResource, setDefaultInterpolationPattern } from 'plain-api';

setDefaultInterpolationPattern(/\:(\w+)/gi);
...
...
const fetchChatMembers = createResource('get', 'https://api.example.com/chat/:chatId/members');
...
...
const chatId = 5;
const members = await fetchChatMembers.call({ chatId });
```
Now `:chatId` will be replaced with `5`.

Other option to override the interpolation pattern for a specific resource is to provide it as an option when creating the resource:
```javascript
import { createResource } from 'plain-api';

const fetchChatMembers = createResource('get', 'https://api.example.com/chat/:chatId/members', {
    interpolationPattern: /\:(\w+)/gi
});
...
...
const chatId = 5;
const members = await fetchChatMembers.call({ chatId });
```
This will override the default interpolation pattern only for that specified resource.

### withCredentials

A boolean indicates whether or not cross-site requests should be made using credentials such as cookies, authorization headers or TLS client certificates. Default is `false`. You can read more [here](https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/withCredentials).


### Headers

`headersMap` let us to decide which parameter will be passed to the header. For example, sending a request with `X-Auth-Token` header:
```javascript
import { createResource } from 'plain-api';

const postMessage = createResource('post', 'https://api.example.com/message', {
    headersMap: {
        token: 'X-Auth-Token'
    }
});
...
...
...
const token = 'this-is-some-user-token';
await postMessage.call({ token });
```
In this example we performed a request and set the header `X-Auth-Token` with the provided token.   
If we call the resource without providing `token`, the header won't be added.


### Body

`inputMap` is used to define the request payload. For example, if our api expects an object in the form `{ user_name, user_age, home_address }`, we will define the following resource:
```javascript
import { createResource } from 'plain-api';

const updateUser = createResource('put', 'https://api.example.com/user', {
    inputMap: {
        name: 'user_name',
        age: 'user_age',
        address: 'home_address'
    }
});
...
...
...
await updateUser.call({
    name: 'Dan',
    age: 23,
    address: 'Somewhere under the sea'
});
```
Parameters that will not be defined in `inputMap` won't be added to the request body.   
Input of `GET` requests is passed using query string.


### Parse the Response

We can define `parsers` array in order to parse the response body. Each parser is a method that gets the parsed response body, a boolean indicator whether the request status code represents a failure and the original payload sent to the request.   
For example:
```javascript
import { createResource } from 'plain-api';

const getUser = createResource('get', 'https://api.example.com/users/{{userId}}', {
    parsers: [
        (data, isFailure, payload) => {
            if (isFailure) {
                throw new Error(`Fail to call api with userId equals to ${payload.userId}`);
            }
            return data.profile;
        },
        profile => {
            name: profile.user_name,
            age: profile.user_age,
            address: profile.home_address
        }
    ]
});
...
...
...
try {
    const user = await getUser.call({ userId: 12 });
    console.log(`Name: ${user.name}, Age: ${user.age}, Address: ${user.address}`);
} catch (err) {
    console.log(`Request failed: ${err.message}`);
}
```
In this example we provide dtwo parsers. If the request failed (status code different from 2xx), the first parser will throw an error. Otherwise it will return the user profile which will be parsed by the second parser.

### Errors handling

* If an api call respond with 2xx status code, everything is fine and no error will be thrown.   
* If an api call respond with failure status code and contains a response (there was a host that got the request and sent a response), no error will be thrown but the parsers will get `true` value in `isFailure`. In this case a parser can decide to throw an error which will be propagate to the api caller.
* If an api call respond with failure status code and doesn't contain a response (there was nobody on the other side, no handler / no server / no internet connection / ... no response), an error will be thrown and no parser will be called.

## Tests

Tests are written with [jest](https://facebook.github.io/jest/). In order to run it, clone this repository and:

```sh
npm test
```

## Release History

* 1.0.0
    * First stable tested version

## Meta

Naor Ye – naorye@gmail.com

Distributed under the MIT license. See ``LICENSE`` for more information.

[https://github.com/naorye/plain-api](https://github.com/naorye/plain-api/)

## Contributing

1. Fork it (<https://github.com/naorye/plain-api/fork>)
2. Create your feature branch (`git checkout -b feature/fooBar`)
3. Commit your changes (`git commit -am 'Add some fooBar'`)
4. Push to the branch (`git push origin feature/fooBar`)
5. Create a new Pull Request

<!-- Markdown link & img dfn's -->
[npm-image]: https://img.shields.io/npm/v/plain-api.svg?style=flat-square
[npm-url]: https://npmjs.org/package/plain-api
[npm-downloads]: https://img.shields.io/npm/dm/plain-api.svg?style=flat-square
[travis-image]: https://img.shields.io/travis/naorye/plain-api/master.svg?style=flat-square
[travis-url]: https://travis-ci.org/naorye/plain-api
[coveralls-image]: https://img.shields.io/coveralls/naorye/plain-api.svg?style=flat-square
[coveralls-url]: https://coveralls.io/github/naorye/plain-api?branch=master
